SystemOrganization addCategory: #TelnetSessionBasic!

SharedPool subclass: #TelnetCmdMatchLib
	instanceVariableNames: ''
	classVariableNames: 'CiscoLogin CiscoOpPrompt CiscoPassword CmdMatch JunCustomPrompt JunLogin JunOpPrompt JunPassword'
	poolDictionaries: ''
	category: 'TelnetSessionBasic'!

!TelnetCmdMatchLib commentStamp: '<historical>' prior: 0!
Strings to match against CLI prompts of network equipment!

!TelnetCmdMatchLib class methodsFor: 'initialization' stamp: 'DanilOsipchuk 6/18/2013 10:56'!
initialize
	"self initialize"
	
	JunLogin := '*login:*'.
	JunPassword := '*password:*'.
	JunOpPrompt := '*>*'.
	JunCustomPrompt := '````TelnetSessionBasic~~>'.
	CiscoLogin := 'Username:*'.
	CiscoPassword := 'Password:*'.
	CiscoOpPrompt := '*>'.
	! !

SharedPool subclass: #TelnetConstantsRFC854
	instanceVariableNames: ''
	classVariableNames: 'AO AYT BRK DM DO DONT EC EL GA IAC IP NOP SB SE SGA WILL WONT'
	poolDictionaries: ''
	category: 'TelnetSessionBasic'!

!TelnetConstantsRFC854 commentStamp: '<historical>' prior: 0!
Straight from RFC854

NAME               CODE              MEANING

      SE                  240    End of subnegotiation parameters.
      NOP                 241    No operation.
      Data Mark           242    The data stream portion of a Synch.
                                 This should always be accompanied
                                 by a TCP Urgent notification.
      Break               243    NVT character BRK.
      Interrupt Process   244    The function IP.
      Abort output        245    The function AO.
      Are You There       246    The function AYT.
      Erase character     247    The function EC.
      Erase Line          248    The function EL.
      Go ahead            249    The GA signal.
      SB                  250    Indicates that what follows is
                                 subnegotiation of the indicated
                                 option.
      WILL (option code)  251    Indicates the desire to begin
                                 performing, or confirmation that
                                 you are now performing, the
                                 indicated option.
      WON'T (option code) 252    Indicates the refusal to perform,
                                 or continue performing, the
                                 indicated option.
      DO (option code)    253    Indicates the request that the
                                 other party perform, or
                                 confirmation that you are expecting
                                 the other party to perform, the
                                 indicated option.
      DON'T (option code) 254    Indicates the demand that the
                                 other party stop performing,
                                 or confirmation that you are no
                                 longer expecting the other party
                                 to perform, the indicated option.
      IAC                 255    Data Byte 255.
!

!TelnetConstantsRFC854 class methodsFor: 'initialization' stamp: 'DanilOsipchuk 7/5/2013 13:05'!
initialize
	"self initialize"
	SGA :=3. "suppress go ahead"
	SE := 240.
	NOP := 241.
	DM := 242.
	BRK := 243.
	IP := 244.
	AO := 245.
	AYT := 246.
	EC := 247.
	EL := 248.
	GA := 249.
	SB := 250.
	WILL := 251.
	WONT := 252.
	DO := 253.
	DONT := 254.
	IAC := 255! !

Object subclass: #TelnetSessionBasic
	instanceVariableNames: 'host port login password socket rawMode'
	classVariableNames: ''
	poolDictionaries: 'TelnetCmdMatchLib TelnetConstantsRFC854'
	category: 'TelnetSessionBasic'!

!TelnetSessionBasic commentStamp: '<historical>' prior: 0!
TelnetSessionBasic is intended for simple automation tasks using telnet protocol.

It provides some convenience remotely comparable to  'expect' scripting command/libraries (#readToMatch:, #cmd:endMatch:).

The implementation uses Socket and hides telnet control flow details (rfc858) from user. Basically it agrees to do SUPPRESS-GO-AHEAD and rejects every other option suggested by remote side, which is barely enough to talk to popular network equipment (see #processReadData:). 






	!

!TelnetSessionBasic class methodsFor: 'examples' stamp: 'DanilOsipchuk 6/18/2013 11:25'!
exampleCiscoLogin
	"self exampleCiscoLogin"
	| ts |
	ts := TelnetSessionBasic new
	setLoginString: 'user:password@host'; "point to remote system"
	open; "open telnet connection"
	telnetLoginCisco; 
	yourself."perform login over opened telnet connection using preconfigured scenario"
	Transcript open.
	Transcript nextPutAll: (ts cmd: 'show users' endMatch: CiscoOpPrompt ); flush. "get output of command, stop at 	the cmd prompt"
	ts close.! !

!TelnetSessionBasic class methodsFor: 'examples' stamp: 'DanilOsipchuk 6/18/2013 11:25'!
exampleJuniperLogin
	"self exampleJuniperLogin"
	| ts |
	ts := TelnetSessionBasic new
	setLoginString: 'user:password@host'; "point to remote system"
	open; "open telnet connection"
	telnetLoginJunos."perform login over opened telnet connection using preconfigured scenario"
	Transcript open.
	Transcript nextPutAll: (ts cmd: 'show route extensive' endMatch: JunCustomPrompt); flush. "get output of command, stop at 	the cmd prompt"
	ts close.! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 7/4/2013 19:39'!
close

	socket
		ifNotNil: 
			[socket close.
			socket := nil]! !

!TelnetSessionBasic methodsFor: 'scripting' stamp: 'DanilOsipchuk 6/18/2013 11:17'!
cmd: aString endMatch: aPattern 
	
	"send aString as a command and read output ending by line matching aPattern"
	self writeln: aString.
	^(self readToMatch: aPattern) allButFirst: aString size + 2  "skip the echoed command in the beginning + crlf "! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 7/5/2013 11:41'!
escapeCollection: aCollection

	|  es  |
	es := WriteStream on: (aCollection class new: aCollection size).
	aCollection do: [:byte |
			byte asInteger == IAC ifTrue: [ es nextPut: byte ].
			es nextPut: byte].
	^es contents! !

!TelnetSessionBasic methodsFor: 'initialize-release' stamp: 'DanilOsipchuk 7/5/2013 11:43'!
initialize 

	self rawModeOff! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 7/5/2013 10:54'!
nextFrom: aStream

	| byte |
	byte  := aStream next.
	(byte asInteger == IAC)
		ifTrue: 
			[| cmd option |
			cmd := aStream next.
			cmd asInteger == IAC 
				ifTrue: [^cmd]
				ifFalse: 
					[option := aStream next.
					self processCmd: cmd option: option].
			^nil ].

	(byte asInteger == 0) ifTrue:  [^nil]. "as we are in a text mode - should we skip zeroes which confuse upper protocol? on not? let us skip"
	^byte   
! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 7/5/2013 11:43'!
open

	"open a socket for a configured session"
	self close.
	socket := Socket new connectToHostNamed: host port: port.
! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 7/5/2013 11:21'!
processCmd: cmd option: option 
	| reply |
	" agree doing SGA, reject everything else"
	reply := nil.
	option asInteger = SGA 
		ifTrue: 
			["suppress_go_ahead suggested - agree"
			cmd asInteger = DO ifTrue: [reply := ByteArray with: IAC with: WILL with: SGA ]. 
			cmd asInteger = WILL  ifTrue: [reply := ByteArray with: IAC with: DO with: SGA ]]
		ifFalse: 
			[cmd asInteger = DO "refuse"
				ifTrue: [reply :=  ByteArray with: IAC with: WONT with: option asInteger].
			cmd asInteger = WILL "reject"
				ifTrue:  [reply := ByteArray with: IAC with: DONT with: option asInteger].
			((cmd asInteger = WONT) or: [cmd asInteger = DONT]) ifTrue: [ ^self "silently agree" ]].
	reply ifNotNil: [socket  sendData: reply]. 
	
	! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 7/5/2013 11:56'!
processReadData: aDataFromSocket 


	| rs ws |
	rs := ReadStream on: aDataFromSocket.
	ws := WriteStream on: (aDataFromSocket class new: aDataFromSocket size).
	[rs atEnd] whileFalse: 
			[| char |
			char := self nextFrom: rs.
			char ifNotNil: [ws nextPut: char] ].
	^ws contents! !

!TelnetSessionBasic methodsFor: 'accessing' stamp: 'DanilOsipchuk 7/5/2013 11:22'!
rawModeOff

	rawMode := false! !

!TelnetSessionBasic methodsFor: 'accessing' stamp: 'DanilOsipchuk 7/5/2013 11:22'!
rawModeOn

	rawMode := true! !

!TelnetSessionBasic methodsFor: 'read' stamp: 'DanilOsipchuk 7/5/2013 11:26'!
read
	"read ready output from socket"

	| rawData |
	rawData := socket receiveAvailableData.
	^ rawMode
		ifTrue: [ rawData ]
		ifFalse: [ self processReadData: rawData ]! !

!TelnetSessionBasic methodsFor: 'read' stamp: 'DanilOsipchuk 6/18/2013 11:24'!
readToMatch: aPattern

	^self readToMatchAnywhere: aPattern  waitFor: 3! !

!TelnetSessionBasic methodsFor: 'read' stamp: 'DanilOsipchuk 6/18/2013 11:10'!
readToMatchAnywhere: aPattern waitFor: seconds
	"read data from socket until any line in buffer matches the pattern. 
	 this should be more reliable but a bit slower than #readToMatchLast:waitFor:"

	| currentLines currentContents currentMatchPosition ws timer matchFound |
	timer := seconds * 1000 + Time millisecondClockValue.
	currentLines := #().
	currentContents := ''.
	ws := WriteStream on: (String new: 64).
	matchFound := false.
	[ matchFound ]
		whileFalse: [ 
			| newData |
			Time millisecondClockValue > timer
				ifTrue: [ self error: 'Timeout' ].
			newData := self read.
			newData isEmpty
				ifFalse: [ "newData arrived, update buffer and look for match"
					| lineIndex |
					ws nextPutAll: newData.
					currentContents := ws contents.
					currentLines := currentContents findTokens: String crlf keep: String crlf.
					currentMatchPosition := 0.	"count from end"
					lineIndex := currentLines size.	"look for a match, scan from the end"
					[ matchFound or: [ lineIndex = 0 ] ]
						whileFalse: [ 
							currentMatchPosition := currentMatchPosition + (currentLines at: lineIndex) size.
							(aPattern match: (currentLines at: lineIndex))
								ifTrue: [ matchFound := true ]
								ifFalse: [ lineIndex := lineIndex - 1 ] ].
					Processor yield ] ].	"we can get here only if match did happen, currentMatchPosition is a sum of lengths of strings beginning from match"
	^ currentContents copyFrom: 1 to: currentContents size - currentMatchPosition! !

!TelnetSessionBasic methodsFor: 'read' stamp: 'DanilOsipchuk 6/18/2013 11:16'!
readToMatchLast: aPattern waitFor: seconds
	"read data from socket until last line in buffer matches the pattern. "

	| lastLine currentContents ws timer  |
	timer := seconds * 1000 + Time millisecondClockValue.
	currentContents := ''. lastLine := ''.
	ws := WriteStream on: (String new: 64).
	[ aPattern match: lastLine ]
		whileFalse: [ 
			| newData |
			Time millisecondClockValue > timer
				ifTrue: [ self error: 'Timeout' ].
			newData := self read.
			newData isEmpty
				ifFalse: [ "newData arrived, update buffer and look for match"
					ws nextPutAll: newData.
					currentContents := ws contents.
					lastLine := currentContents lines last.
					Processor yield ] ].	
	^ currentContents copyFrom: 1 to: currentContents size - lastLine size! !

!TelnetSessionBasic methodsFor: 'scripting' stamp: 'DanilOsipchuk 7/5/2013 12:54'!
script: aString endMatch: aPattern 
	
	"send each line as a command and return output of last command. Output of each command should match aPattern"
	| output |
	output := ''.
	aString linesDo: [ :line | output := self cmd: line endMatch: aPattern ].
	^output! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 7/2/2013 18:36'!
setLoginString: aString 
	"set login credentials aString:   login:password@host:port "

	| tmp lp hp |
	tmp := aString subStrings: '@'.
	tmp size = 2 
		ifTrue: 
			[lp := tmp first subStrings: ':'.
			hp := tmp second subStrings: ':'.
			login := lp first.
			lp size = 2 ifTrue: [password := lp last]]
		ifFalse: [hp := tmp first subStrings: ':'].	"both login and host - extract user/password"

	"common part: parse the host port pair"
	port := 23.
	hp size = 2 ifTrue: [port := hp second asNumber].
	host := hp first! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 7/5/2013 13:13'!
telnetLoginCisco

	"remote system is Cisco, send login and password (with username, without enable)."


	self readToMatch: CiscoLogin. self writeln: login. 
	self readToMatch: CiscoPassword. self writeln: password.
	self readToMatch: CiscoOpPrompt.
	self cmd: 'terminal length 0' endMatch: CiscoOpPrompt.
	self cmd: 'terminal width 512' endMatch: CiscoOpPrompt.
! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 6/18/2013 11:02'!
telnetLoginJunos

	"remote system is JUNOS, send login and password"

	self readToMatch: JunLogin. self writeln: login. 
	self readToMatch: JunPassword. self writeln: password.
	self readToMatch: JunOpPrompt.
	self cmd: ('set cli prompt ' , JunCustomPrompt) endMatch: JunCustomPrompt. "set a prompt which can be efficiently and reliably matched against"
	self cmd: 'set cli screen-length 0' endMatch: JunCustomPrompt.
	self cmd: 'set cli screen-width 0' endMatch: JunCustomPrompt.

	
! !

!TelnetSessionBasic methodsFor: 'write' stamp: 'DanilOsipchuk 7/5/2013 11:27'!
write: aString
	"send a string to remote system"

	rawMode
		ifTrue: [ socket sendData: aString ]
		ifFalse: [ socket sendData: (self escapeCollection: aString) ]! !

!TelnetSessionBasic methodsFor: 'write' stamp: 'DanilOsipchuk 7/5/2013 11:28'!
writeln: aString

	self write: (aString , String crlf)! !
TelnetCmdMatchLib initialize!
TelnetConstantsRFC854 initialize!
